package main

// ctr wide constants
const (
	// ExitStatusOK indicates successful completion
	ExitStatusOK = 0

	// ExitStatusMissingArg indicates failure due to missing argument(s)
	ExitStatusMissingArg = 1
)
